/*
 * PWCBSD - Philips USB webcam driver for FreeBSD 5.4 and higher
 *
 *  Copyright (C) 2006 Raaf 
 *
 * Based on the Linux pwc driver.
 *  
 *  Copyright (C) 1999-2003 Nemosoft Unv.
 *  Copyright (C) 2004-2006 Luc Saillard
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */

#ifndef PWC_H
#define PWC_H
#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: pwc.c, v 1.00 2008/02/23 18:00:00 vha $");
#include <sys/param.h>
#include <sys/select.h>
#include <sys/systm.h>
#include <sys/poll.h>
#include <sys/proc.h>
#include <sys/kernel.h>
#include <sys/fcntl.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/vnode.h>
#include <sys/syslog.h>
#include <sys/selinfo.h>
#include <i386/pmap.h>
#include <sys/lkm.h>
#include <sys/malloc.h>

#include <machine/vmparam.h>	/* PAGE_SIZE */

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usbdevs.h>
#include <dev/usb/usb_quirks.h>

#include "pwc-uncompress.h"
#include "pwc-ioctl.h"
#include "videodev.h"

#define pwc_device		upwc_softc
#define PWC_ASLEEP		1
#define PWC_INIT		2
#define PWC_POLL		4

/* Trace certain actions in the driver */
#define TRACE_MODULE		0x0001
#define TRACE_PROBE		0x0002
#define TRACE_OPEN		0x0004
#define TRACE_READ		0x0008
#define TRACE_MEMORY		0x0010
#define TRACE_FLOW		0x0020
#define TRACE_SIZE		0x0040
#define TRACE_PWCX		0x0080
#define TRACE_IOCTL		0x0100
#define TRACE_ISOC		0x0200
#define TRACE_ISOC_VERBOSE	0x0400
#define TRACE_READ_VERBOSE	0x0800
#define TRACE_SEQUENCE		0x1000

#ifdef USB_DEBUG
#define Trace(R, A...)		if (pwcdebug & R) printf(PWC_NAME " " A)
#define Debug(A...)		if(pwcdebug) printf(PWC_NAME " " A)
extern int  pwcdebug;
#else
#define Trace(R, A...)
#define Debug(A...)
#endif

#define Info(A...)		printf(PWC_NAME " " A)
#define Err(A...)		printf(PWC_NAME " " A)


/* Defines for ToUCam cameras */
#define TOUCAM_HEADER_SIZE	8
#define TOUCAM_TRAILER_SIZE	4

#define FEATURE_MOTOR_PANTILT	0x0001

#define PWC_NAME		"pwc"

/* Ignore errors in the first N frames, to allow for startup delays */
#define FRAME_LOWMARK		5

/* Size and number of buffers for the ISO pipe. */
#define MAX_ISO_BUFS		3

/* !!!!!!!!!! IMPORTANT DO NOT SET THIS HIGHER THAN 8 !!!!!!!!!!!!!!! 
 * !!!!!!!!!! OHCI IS BUGGY IF YOU DO SO              !!!!!!!!!!!!!!!*/
#define ISO_FRAMES_PER_DESC	8

#define ISO_MAX_FRAME_SIZE	960
#define ISO_BUFFER_SIZE 	(ISO_FRAMES_PER_DESC * ISO_MAX_FRAME_SIZE)

/* Frame buffers: contains compressed or uncompressed video data. */
#define MAX_FRAMES		5
/* Maximum size after decompression is 640x480 YUV data, 1.5 * 640 * 480 */
#define PWC_FRAME_SIZE 		(460800 + TOUCAM_HEADER_SIZE + TOUCAM_TRAILER_SIZE)

/* Absolute maximum number of buffers available for mmap() */
#define MAX_IMAGES 		10

#ifndef V4L2_PIX_FMT_PWC1
#define V4L2_PIX_FMT_PWC1	v4l2_fourcc('P','W','C','1')
#define V4L2_PIX_FMT_PWC2	v4l2_fourcc('P','W','C','2')
#endif

/* intermediate buffers with raw data from the USB cam */
struct pwc_frame_buf
{
   char *		  data;
   volatile int		  filled;	    /* number of bytes filled */
   struct pwc_frame_buf * next;		    /* list */
};

/* additional information used in dealing image between kernel and userland */
struct pwc_imgbuf
{
  char  *		  bufmem;	    /* addr of buffer in kernel space */
  int			  vma_use_count;    /* # of time memory is mapped */
};

struct upwc_softc;

struct pwc_iso_buf
{
  unsigned char *	  data;
  usbd_xfer_handle	  xfer;
  u_int16_t		  sizes [ISO_FRAMES_PER_DESC];
  struct upwc_softc *	  sc;
};
 
struct upwc_softc
{
   USBBASEDEVICE	  sc_dev;	    /* Base device */
   usbd_device_handle	  udev;		    /* USB device */
   usbd_interface_handle  sc_iface;	    /* Data interface */
   usbd_pipe_handle	  sc_videopipe;	    /* Video pipe */
   struct cdev *	  sc_dev_t;
   int			  maj;		    /* Major */
   
   struct selinfo	  rsel;
   char *		  name;
   int			  type;		    /* type of cam */
   int			  release;	    /* release number */
   int			  features;	    /* feature bits */
   char			  serial [USB_MAX_STRING_LEN]; /* serial number */
   int			  error_status;	    /* something wrong with the cam*/
   int			  usb_init;	    /* the cam is initialized */

   int			  power_save;
   int			  led_on;
   int			  led_off;
   int			  stats;
   int			  pwc_mbufs;
   int			  pwc_fbufs;
   int			  pwc_pad;

   /*** Video data ***/
   int			  vopen;	    /* flag */
   int			  vendpoint;	    /* video isoc endpoint */
   int			  vcinterface;	    /* video control interface */
   int			  valternate;	    /* alternate interface needed */
   int			  vframes, vsize;   /* FPS & size (see PSZ_*) */
   int			  vpalette;	    /* palette: 420P, RAW or RGBBAYER */
   int			  vframe_count;	    /* received frames */
   int			  vframes_dumped;   /* counter for dumped frames */
   int			  vframes_error;    /* frames received in error */
   int			  vmax_packet_size; /* USB maxpacket size */
   u_int16_t		  vlast_packet_size;/* for frame synchronisation */
   int			  visoc_errors;	    /* # of contiguous ISOC errors */
   int			  vcompression;	    /* desired compression factor */
   int			  vbandlength;	    /* compressed band length; 0 = no */
   char			  vsnapshot;	    /* snapshot mode */
   char			  vsync;	    /* used by isoc handler */
   char			  vmirror;	    /* for ToUCaM series */
   
   int			  cmd_len;
   unsigned char	  cmd_buf [13];

   /* The image acquisition requires 3 to 4 steps:
      1. data is gathered in short packets from the USB controller
      2. data is synchronized and packed into a frame buffer
      3a. in case data is compressed, decompress it directly into image buffer
      3b. in case data is uncompressed, copy into image buffer with viewport
      4. data is transferred to the user process

      Note that MAX_ISO_BUFS != MAX_FRAMES != MAX_IMAGES....
      We have in effect a back-to-back-double-buffer system.
    */
   /* 1: isoc */
   struct pwc_iso_buf	  sbuf [MAX_ISO_BUFS];
   char			  iso_init;

   /* 2: frame */
   struct pwc_frame_buf * fbuf;		    /* all frames */
   struct pwc_frame_buf * empty_frames, *empty_frames_tail; /* empty frames */
   struct pwc_frame_buf * full_frames, *full_frames_tail;   /* filled frames */
   struct pwc_frame_buf * fill_frame;	    /* frame currently being filled */
   struct pwc_frame_buf * read_frame;	    /* frame currently read */
   int			  frame_header_size, frame_trailer_size;
   int			  frame_size;
   int			  frame_total_size; /* including header & trailer */
   int			  drop_frames;

   /* 3: decompression */
   void *		  decompress_data;  /* private data for decompression */

   /* 4: image */
   /* We have an 'image' and a 'view', where 'image' is the fixed-size image
      as delivered by the camera, and 'view' is the size requested by the
      program. The camera image is centered in this viewport, laced with
      a gray or black border. view_min <= image <= view <= view_max;
    */
   int			  image_mask;	    /* bitmask of supported sizes */
   struct pwc_coord	  view_min, view_max; /* min and max viewable sizes */
   struct pwc_coord	  abs_max;          /* max size with compression */
   struct pwc_coord	  image, view;	    /* image and viewport size */
   struct pwc_coord	  offset;	    /* offset within the viewport */

   void *		  image_data;	    /* total buffer subdivided in... */
   struct pwc_imgbuf	  images [MAX_IMAGES]; /* ...several images... */
   int			  fill_image;	    /* ...which are rotated. */
   int			  len_per_image;    /* length per image */
   int			  image_read_pos;   /* were we are in the imagebuffer */
   int			  image_used [MAX_IMAGES];  /* For MCAPTURE and SYNC */

   kmutex_t		  ptrlock;	    /* for buffer pointers */
   
   /*** motorized pan/tilt feature */
   struct pwc_mpt_range	  angle_range;
   int			  pan_angle;	    /* in degrees * 100 */
   int			  tilt_angle;	    /* absolute angle; 0 is home */
   int			  snapshot_button_status; 

   /*** Misc. data ***/
   int			  state;	    /* Waiting for a frame to end... */
};

/** functions in pwc-if.c */
int			  pwc_try_video_mode  (struct pwc_device *  pdev, 
					       int		    width, 
					       int		    height, 
					       int		    new_fps, 
					       int		    new_cmp, 
					       int		    new_snap);

int			  pwc_handle_frame    (struct pwc_device *  pdev);
void			  pwc_next_image      (struct pwc_device *  pdev);

/** Functions in pwc-misc.c */
/* sizes in pixels */
extern struct pwc_coord	  pwc_image_sizes [PSZ_MAX];

int			  pwc_decode_size     (struct pwc_device *  pdev, 
					       int		    width,
					       int		    height);

void			  pwc_construct	      (struct pwc_device *  pdev);

/** Functions in pwc-ctrl.c */
/* Request a certain video mode. Returns < 0 if not possible */
extern int		  pwc_set_video_mode  (struct pwc_device *  pdev,
					       int		    width,
					       int		    height,
					       int		    frames,
					       int		    compression,
					       int		    snapshot);

/* Calculate the number of bytes per image (not frame) */
extern int		  pwc_mpt_reset	      (struct pwc_device *  pdev, 
					       int		    flags);

extern int		  pwc_mpt_set_angle   (struct pwc_device *  pdev,
					       int		    pan,
					       int		    tilt);

/* Various controls; should be obvious. Value 0..65535, or < 0 on error */
extern int pwc_get_brightness(struct pwc_device *pdev);
extern int pwc_set_brightness(struct pwc_device *pdev, int value);
extern int pwc_get_contrast(struct pwc_device *pdev);
extern int pwc_set_contrast(struct pwc_device *pdev, int value);
extern int pwc_get_gamma(struct pwc_device *pdev);
extern int pwc_set_gamma(struct pwc_device *pdev, int value);
extern int pwc_get_saturation(struct pwc_device *pdev);
extern int pwc_set_saturation(struct pwc_device *pdev, int value);
extern int pwc_set_leds(struct pwc_device *pdev, int on_value, int off_value);
extern int pwc_get_leds(struct pwc_device *pdev, int *on_value, int *off_value);
extern int pwc_get_cmos_sensor(struct pwc_device *pdev, int *sensor);
extern int pwc_restore_user(struct pwc_device *pdev);
extern int pwc_save_user(struct pwc_device *pdev);
extern int pwc_restore_factory(struct pwc_device *pdev);

/* exported for use by v4l2 controls */
extern int pwc_get_red_gain(struct pwc_device *pdev, int *value);
extern int pwc_set_red_gain(struct pwc_device *pdev, int value);
extern int pwc_get_blue_gain(struct pwc_device *pdev, int *value);
extern int pwc_set_blue_gain(struct pwc_device *pdev, int value);
extern int pwc_get_awb(struct pwc_device *pdev);
extern int pwc_set_awb(struct pwc_device *pdev, int mode);
extern int pwc_set_agc(struct pwc_device *pdev, int mode, int value);
extern int pwc_get_agc(struct pwc_device *pdev, int *value);
extern int pwc_set_shutter_speed(struct pwc_device *pdev, int mode, int value);
extern int pwc_get_shutter_speed(struct pwc_device *pdev,int *value);
extern int pwc_set_colour_mode(struct pwc_device *pdev, int colour);
extern int pwc_get_colour_mode(struct pwc_device *pdev, int *colour);

/* Power down or up the camera; not supported by all models */
extern int pwc_camera_power(struct pwc_device *pdev, int power);

/* Private ioctl()s; see pwc-ioctl.h */
extern int pwc_do_ioctl(struct pwc_device *pdev, unsigned int cmd, void *arg);

/** Functions in pwc-v4l.c */
extern int pwc_video_do_ioctl(struct pwc_device *pdev, unsigned int cmd, void *arg, int unit);
/** pwc-uncompress.c */
/* Expand frame to image, possibly including decompression. Uses read_frame and fill_image */
extern int pwc_decompress(struct pwc_device *pdev);

#endif
